// Copyright(c) 1999 arton
#include "stdafx.h"
#include "MacCapV2.h"
#include "FilterParam.h"
#include "MacCapture.h"
/////////////////////////////////////////////////////////////////////////////
// CFilterParam

HRESULT STDMETHODCALLTYPE CFilterParam::put_Value( 
            /* [in] */ VARIANT NewVal)
{
	Lock();
	VariantClear(&m_vValue);
	if (NewVal.vt == (VT_VARIANT | VT_BYREF))
	{
		VariantCopy(&m_vValue, NewVal.pvarVal);
	}
	else
	{
		VariantCopy(&m_vValue, &NewVal);
	}
	Unlock();
	return S_OK;
}
        
HRESULT STDMETHODCALLTYPE CFilterParam::get_Value( 
		    /* [retval][out] */ VARIANT __RPC_FAR *pResult)
{
	if (!pResult) return E_POINTER;
	VariantInit(pResult);
	Lock();
	VariantCopy(pResult, &m_vValue);
	Unlock();
	return S_OK;
}
        
HRESULT STDMETHODCALLTYPE CFilterParam::put_Source( 
            /* [in] */ VARIANT_BOOL NewVal)
{
	m_fSrc = (NewVal) ? true : false;
	return S_OK;
}
        
HRESULT STDMETHODCALLTYPE CFilterParam::get_Source( 
            /* [retval][out] */ VARIANT_BOOL __RPC_FAR *pResult)
{
	if (!pResult) return E_POINTER;
	*pResult = (m_fSrc) ? VARIANT_TRUE : VARIANT_FALSE;
	return S_OK;
}
        
HRESULT STDMETHODCALLTYPE CFilterParam::put_Destination( 
            /* [in] */ VARIANT_BOOL NewVal)
{
	m_fDst = (NewVal) ? true : false;
	return S_OK;
}
        
HRESULT STDMETHODCALLTYPE CFilterParam::get_Destination( 
            /* [retval][out] */ VARIANT_BOOL __RPC_FAR *pResult)
{
	if (!pResult) return E_POINTER;
	*pResult = (m_fDst) ? VARIANT_TRUE : VARIANT_FALSE;
	return S_OK;
}
        
HRESULT STDMETHODCALLTYPE CFilterParam::put_Type( 
            /* [in] */ FilterTypeConstants NewVal)
{
	if ((NewVal >= FT_MACADDRESS && NewVal <= FT_UDPPORT)
		|| (NewVal >= FT_JOINT_AND && NewVal <= FT_NOT)) 
	{
		m_ftype = NewVal;
		return S_OK;
	}
	return Error(IDS_E_BADPARAM);
}
        
HRESULT STDMETHODCALLTYPE CFilterParam::get_Type( 
            /* [retval][out] */ FilterTypeConstants __RPC_FAR *pRetVal)
{
	if (!pRetVal) return E_POINTER;
	*pRetVal = m_ftype;
	return S_OK;
}

HRESULT STDMETHODCALLTYPE CFilterParam::get_Valid( 
            /* [retval][out] */ VARIANT_BOOL __RPC_FAR *pResult)
{
	if (!pResult) return E_POINTER;
	USES_CONVERSION;
	*pResult = VARIANT_FALSE;
	bool fCopyNeed = false;
	VARIANT v;
	VariantInit(&v);
	//
	switch (m_ftype)
	{
	case FT_MACADDRESS:
		v.vt = VT_ARRAY | VT_UI1;
		if (m_vValue.vt == VT_BSTR)
		{
			OLECHAR sz[16];
			int i(0);
			LPOLESTR p = m_vValue.bstrVal;
			for (; *p; p++)
			{
				if (iswxdigit(*p))
				{
					sz[i++] = *p;
				}
			}
			if (i == 12)
			{
				sz[i] = L'\0';
				SAFEARRAYBOUND rgsabound[1];
				rgsabound[0].cElements = 6;
				rgsabound[0].lLbound = 0;
				v.parray = SafeArrayCreate(VT_UI1, 1, rgsabound);
				SafeArrayLock(v.parray);
				Pack(v.parray->pvData, sz, i);
				SafeArrayUnlock(v.parray);
				*pResult = VARIANT_TRUE;
				fCopyNeed = true;
			}
		}
		else if (m_vValue.vt == (VT_ARRAY | VT_UI1))
		{
			SAFEARRAY* pAry = m_vValue.parray;
			if (pAry->rgsabound[0].cElements == 6)
				*pResult = VARIANT_TRUE;
		}
		else if (m_vValue.vt == (VT_BYREF | VT_ARRAY | VT_UI1))
		{
			SAFEARRAY* pAry = *(m_vValue.pparray);
			if (pAry->rgsabound[0].cElements == 6)
			{
				SafeArrayCopy(pAry, &(v.parray));
				*pResult = VARIANT_TRUE;
				fCopyNeed = true;
			}
		}
		break;
	case FT_IPADDRESS:
		v.vt = VT_I4;
		if (m_vValue.vt == VT_BSTR)
		{
			v.lVal = (long)inet_addr(W2A(m_vValue.bstrVal));
			if (v.lVal != INADDR_NONE)
			{
				*pResult = VARIANT_TRUE;
				fCopyNeed = true;
			}
		}
		else if (m_vValue.vt == VT_I4)
		{
			*pResult = VARIANT_TRUE;
		}
		break;
	case FT_TCPPORT:
	case FT_UDPPORT:
	case FT_ETHERTYPE:
		if (VariantChangeType(&v, &m_vValue, 0, VT_I2) == S_OK)
		{
			*pResult = VARIANT_TRUE;
			fCopyNeed = true;
		}
		break;
	case FT_PROTOCOL:
		if (VariantChangeType(&v, &m_vValue, 0, VT_UI1) == S_OK)
		{
			*pResult = VARIANT_TRUE;
			fCopyNeed = true;
		}
		break;
	case FT_PACKETOFFSETVALUE:
		{
			SAFEARRAY* pAry = NULL;
			if (m_vValue.vt == (VT_ARRAY | VT_VARIANT))
			{
				pAry = m_vValue.parray;
			}
			else if (m_vValue.vt == (VT_BYREF | VT_ARRAY | VT_VARIANT))
			{
				pAry = *(m_vValue.pparray);
			}
			if (pAry && pAry->rgsabound[0].cElements == 2)
			{
				VARIANT* pVar = (VARIANT*)pAry->pvData;
				VARIANT av[2];
				for (int i = 0; i < 2; i++) VariantInit(&av[i]);
				if (VariantChangeType(&av[0], pVar, 0, VT_I2) == S_OK)
				{
					VariantClear(pVar);
					*pVar = av[0];
					pVar++;
					if (pVar->vt == VT_BSTR)
					{
						int len = SysStringLen(pVar->bstrVal);
						*pResult = VARIANT_TRUE;
						for (int i = 0; i < len; i++)
						{
							if (!iswxdigit(*(pVar->bstrVal + i)))
							{
								*pResult = VARIANT_FALSE;
								break;
							}
						}
					}
				}
			}
		}
		break;
	case FT_JOINT_AND:
	case FT_JOINT_OR:
		{
			SAFEARRAY* pAry = NULL;
			if (m_vValue.vt == (VT_ARRAY | VT_DISPATCH))
			{
				pAry = m_vValue.parray;
			}
			else if (m_vValue.vt == (VT_BYREF | VT_ARRAY | VT_DISPATCH))
			{
				pAry = *(m_vValue.pparray);
			}
			if (pAry && pAry->rgsabound[0].cElements >= 2)
			{
				SafeArrayLock(pAry);
				LPDISPATCH* pDisp = reinterpret_cast<LPDISPATCH*>(pAry->pvData);
				IFilterParam* pTemp;
				*pResult = VARIANT_TRUE;
				for (int i = 0; i < pAry->rgsabound[0].cElements; i++, pDisp++)
				{
					HRESULT hr = (*pDisp)->QueryInterface(IID_IFilterParam, (void**)&pTemp);
					if (hr == S_OK)
					{
						pTemp->Release();
					}
					else
					{
						*pResult = VARIANT_FALSE;
						break;
					}
				}
				SafeArrayUnlock(pAry);
			}
			else if (m_vValue.vt == (VT_ARRAY | VT_VARIANT) && m_vValue.parray->rgsabound[0].cElements >= 2)
			{
				SafeArrayLock(m_vValue.parray);
				VARIANT* pv = reinterpret_cast<VARIANT*>(m_vValue.parray->pvData);
				IFilterParam* pTemp;
				*pResult = VARIANT_TRUE;
				for (int i = 0; i < m_vValue.parray->rgsabound[0].cElements; i++, pv++)
				{
					HRESULT hr = pv->pdispVal->QueryInterface(IID_IFilterParam, (void**)&pTemp);
					if (hr == S_OK)
					{
						pTemp->Release();
					}
					else
					{
						*pResult = VARIANT_FALSE;
						break;
					}
				}
				if (*pResult == VARIANT_TRUE)
				{
					pv = reinterpret_cast<VARIANT*>(m_vValue.parray->pvData);
					SAFEARRAYBOUND rgsabound[1];
					rgsabound[0].cElements = m_vValue.parray->rgsabound[0].cElements;
					rgsabound[0].lLbound = 0;
					v.vt = VT_DISPATCH | VT_ARRAY;
					v.parray = SafeArrayCreate(VT_DISPATCH, 1, rgsabound);
					SafeArrayLock(v.parray);
					IDispatch** pDisp = reinterpret_cast<IDispatch**>(v.parray->pvData);
					for (i = 0; i < m_vValue.parray->rgsabound[0].cElements; i++, pDisp++, pv++)
					{
						pv->pdispVal->QueryInterface(IID_IFilterParam, (void**)pDisp);
					}
					SafeArrayUnlock(v.parray);
					fCopyNeed = true;
				}
				SafeArrayUnlock(m_vValue.parray);
			}
		}
		break;
	case FT_NOT:
		if (m_vValue.vt == VT_DISPATCH)
		{
			IFilterParam* pTemp;
			HRESULT hr = m_vValue.pdispVal->QueryInterface(IID_IFilterParam, (void**)&pTemp);
			if (hr == S_OK)
			{
				pTemp->Release();
				*pResult = VARIANT_TRUE;
			}
		}
		break;
	default:
		break;
	}
	if (fCopyNeed)
	{
		VariantClear(&m_vValue);
		m_vValue = v;
	}
	return S_OK;
}
