//---------------------------------------------------------------------------
// Copyright(c) 1999 arton
//
#include <vcl.h>
#include "activscp.h"
#include <Clipbrd.hpp>
#pragma hdrstop

#include "Filterdoc.h"
#include "MSXML_TLB.h"
#include "FreePeekMain.h"

//---------------------------------------------------------------------------
#pragma package(smart_init)

const char* TFilterDocument::tagtype[] = {
    "address",
    "etype",
    "ip",
    "value",
    "and",
    "or",
    "not",
};

TFilterDocument::TFilterDocument(AnsiString& filename)
    : ForClipboard(false)
{
    fstr = new TFileStream(filename, fmCreate | fmShareDenyWrite);
    bstrValue = bstrType = bstrDirection = bstrOffset = bstrData = NULL;
}

TFilterDocument::TFilterDocument(AnsiString& filename, TCOMIXMLDOMDocument& dom)
    : fstr(NULL), domDoc(dom), ForClipboard(false)
{
    TOLEBOOL success;
    domDoc.load(TVariant(filename), &success);
    fSuccess = success;
    InitBstrs();
}

TFilterDocument::TFilterDocument(TCOMIXMLDOMDocument& dom)
    : fstr(NULL), domDoc(dom), ForClipboard(false)
{
    BSTR bstr = StringToOleStr(Clipboard()->AsText);
    fSuccess = (domDoc.loadXML(bstr)) ? true : false;
    SysFreeString(bstr);
    InitBstrs();
}

TFilterDocument::TFilterDocument()
    : fstr(NULL), ForClipboard(true)
{
    bstrValue = bstrType = bstrDirection = bstrOffset = bstrData = NULL;
    Clipboard()->Open();
}

void TFilterDocument::InitBstrs()
{
    bstrValue = SysAllocString(L"value");
    bstrType = SysAllocString(L"type");
    bstrDirection = SysAllocString(L"direction");
    bstrOffset = SysAllocString(L"offset");
    bstrData = SysAllocString(L"data");
}

bool TFilterDocument::IsValid()
{
    if (fSuccess)
    {
        IXMLDOMElementPtr root = domDoc.get_documentElement();
        if (root)
        {
            BSTR bstr = root->tagName;
            if (wcscmp(bstr, L"filter"))
                fSuccess = false;
            SysFreeString(bstr);
        }
    }
    return fSuccess;
}

//---------------------------------------------------------------------------
TFilterDocument::~TFilterDocument()
{
    if (ForClipboard)
    {
        Clipboard()->Close();
    }
    if (!domDoc == false)
    {
        IPersistStreamInit* pInit;
        if (domDoc->QueryInterface(IID_IPersistStreamInit, (void**)&pInit) == S_OK)
        {
            pInit->InitNew();   // clear
            pInit->Release();
        }
    }
    delete fstr;
    if (bstrValue) SysFreeString(bstrValue);
    if (bstrType) SysFreeString(bstrType);
    if (bstrDirection) SysFreeString(bstrDirection);
    if (bstrOffset) SysFreeString(bstrOffset);
    if (bstrData) SysFreeString(bstrData);
}
//---------------------------------------------------------------------------
void TFilterDocument::Load(TTreeNodes* Nodes)
{
    IXMLDOMElementPtr root = domDoc.get_documentElement();
    if (root)
    {
        BSTR bstr = root->tagName;
        if (wcscmp(bstr, L"filter") == 0)
        {
            IXMLDOMNodeListPtr list = root->childNodes;
            IXMLDOMNodePtr node = list->nextNode();
            if (node)
            {
                Nodes->Clear();
                LoadNode(node, Nodes, NULL);
            }
        }
        SysFreeString(bstr);
        Nodes->Owner->FullExpand();
    }
}
//---------------------------------------------------------------------------
void TFilterDocument::Load(TTreeNodes* Nodes, TTreeNode* Node)
{
    IXMLDOMElementPtr root = domDoc.get_documentElement();
    if (root)
    {
        BSTR bstr = root->tagName;
        if (wcscmp(bstr, L"filter") == 0)
        {
            IXMLDOMNodeListPtr list = root->childNodes;
            IXMLDOMNodePtr node = list->nextNode();
            if (node)
            {
                Node->DeleteChildren();
                LoadNode(node, Nodes, Node);
                Node->getFirstChild()->MoveTo(Node, naInsert);
                Node->Delete();
            }
        }
        SysFreeString(bstr);
        Nodes->Owner->FullExpand();
    }
}
//---------------------------------------------------------------------------
int TFilterDocument::GetAttr(::IXMLDOMElement* elem, BSTR& bstr)
{
    int n(-1);
    VARIANT v;
    elem->getAttribute(bstr, (TVariant*)&v);
    if (v.vt == VT_BSTR)
    {
        n = _wtoi(v.bstrVal);
        VariantClear(&v);
    }
    return n;
}
//---------------------------------------------------------------------------
void TFilterDocument::LoadNode(IXMLDOMNodePtr docNode, TTreeNodes* Nodes, TTreeNode* Node)
{
    ::IXMLDOMElement* pElem;
    BSTR bstr;
    if (docNode->QueryInterface(::IID_IXMLDOMElement, (void**)&pElem) == S_OK
        && pElem->get_tagName(&bstr) == S_OK)
    {
        AnsiString atag(bstr);
        SysFreeString(bstr);
        int iData;
        for (int i = 0; i < COUNTOF(tagtype); i++)
        {
            if (atag.AnsiCompare(tagtype[i]) == 0)
            {
                TTreeNode* NewNode;
                bstr = docNode->text;
                iData = 0;
                switch (i + FBT_ADDRESS)
                {
                case FBT_ADDRESS:
                    iData = MAKELONG(MAKEWORD(GetAttr(pElem, bstrType), GetAttr(pElem, bstrDirection)), 0);
                    NewNode = Nodes->AddChildObject(Node, bstr, reinterpret_cast<void*>(iData));
                    break;
                case FBT_ETYPE:
                    iData = GetAttr(pElem, bstrValue);
                    NewNode = Nodes->AddChildObject(Node, bstr, reinterpret_cast<void*>(iData));
                    break;
                case FBT_IP:
                    iData = MAKELONG(MAKEWORD(GetAttr(pElem, bstrType), GetAttr(pElem, bstrDirection)), GetAttr(pElem, bstrValue));
                    NewNode = Nodes->AddChildObject(Node, bstr, reinterpret_cast<void*>(iData));
                    break;
                case FBT_VALUE:
                    iData = (GetAttr(pElem, bstrOffset) < 0) ? -1 : 0;
                    NewNode = Nodes->AddChildObject(Node, bstr, reinterpret_cast<void*>(iData));
                    break;
                case FBT_AND:
                case FBT_OR:
                case FBT_NOT:
                    NewNode = Nodes->AddChild(Node, atag.UpperCase());
                    {
                        IXMLDOMNodeListPtr list = docNode->childNodes;
                        IXMLDOMNodePtr node;
                        while (node = list->nextNode())
                        {
                            LoadNode(node, Nodes, NewNode);
                        }
                    }
                    break;
                }
                NewNode->ImageIndex = i + FBT_ADDRESS;
                NewNode->SelectedIndex = i + FBT_ADDRESS;
                SysFreeString(bstr);
                break;
            }
        }
        pElem->Release();
    }
}
//---------------------------------------------------------------------------
void TFilterDocument::Save(TTreeNode* pRoot)
{
    static const char hdr[] = "<?xml version=\"1.0\" encoding=\"shift_jis\"?>\r\n<filter>\r\n";
    static const char ftr[] = "</filter>\r\n";
    if (ForClipboard)
        Clipboard()->AsText = hdr;
    else
        fstr->Write(hdr, sizeof(hdr) - 1);
    SaveNode(pRoot, 0);
    if (ForClipboard)
        Clipboard()->AsText = Clipboard()->AsText + ftr;
    else
        fstr->Write(ftr, sizeof(ftr) - 1);
}
//---------------------------------------------------------------------------
void TFilterDocument::SaveNode(TTreeNode* pNode, int lev)
{
    AnsiString& sSpaces = AnsiString::StringOfChar(' ', lev);
    char sz[128];
    int cb = wsprintf(sz, "%s<%s", sSpaces.c_str(), tagtype[pNode->ImageIndex - FBT_ADDRESS]);
    if (ForClipboard)
        Clipboard()->AsText = Clipboard()->AsText + sz;
    else
        fstr->Write(sz, cb);
    int val = reinterpret_cast<int>(pNode->Data);
    if (val < 0)
    {
        cb = wsprintf(sz, ">\r\n");
    }
    else
    {
        switch (pNode->ImageIndex)
        {
        case FBT_ADDRESS:
            {
            AnsiString& as = pNode->Text;
            int pos = as.Pos(":");
            cb = wsprintf(sz, " type=\"%d\" direction=\"%d\" value=\"%s\">\r\n%s  %s\r\n",
                LOBYTE(LOWORD(val)), HIBYTE(LOWORD(val)),
                as.SubString(pos + 1, as.Length() - pos).c_str(),
                sSpaces.c_str(), as.c_str());
            }
            break;
        case FBT_ETYPE:
            cb = wsprintf(sz, " value=\"%d\">\r\n%s  %s\r\n", val, sSpaces.c_str(), pNode->Text.c_str());
            break;
        case FBT_IP:
            cb = wsprintf(sz, "  type=\"%d\" direction=\"%d\" value=\"%d\">\r\n%s  %s\r\n",
                LOBYTE(LOWORD(val)), HIBYTE(LOWORD(val)),
                HIWORD(val),
                sSpaces.c_str(), pNode->Text.c_str());
            break;
        case FBT_VALUE:
            {
            AnsiString& as = pNode->Text;
            int pos = as.Pos(":");
            cb = wsprintf(sz, " offset=\"%s\" data=\"%s\">\r\n%s  %s\r\n",
                as.SubString(1, pos - 1).c_str(),
                as.SubString(pos + 1, as.Length() - pos).c_str(),
                sSpaces.c_str(), as.c_str());
            }
            break;
        case FBT_AND:
        case FBT_OR:
        case FBT_NOT:
            cb = wsprintf(sz, ">\r\n");
            break;
        }
    }
    if (ForClipboard)
        Clipboard()->AsText = Clipboard()->AsText + sz;
    else
        fstr->Write(sz, cb);
    TTreeNode* pChild = pNode->getFirstChild();
    while (pChild)
    {
        SaveNode(pChild, lev + 4);
        pChild = pNode->GetNextChild(pChild);
    }
    cb = wsprintf(sz, "%s</%s>\r\n", sSpaces.c_str(), tagtype[pNode->ImageIndex - FBT_ADDRESS]);
    if (ForClipboard)
        Clipboard()->AsText = Clipboard()->AsText + sz;
    else
        fstr->Write(sz, cb);
}

