// Copyright(c) 1999 arton
#include "stdafx.h"
#include "MacCapV2.h"
#include "MacPacket.h"

#include "AssNum.h"

LPOLESTR Unpack(LPOLESTR, LPVOID, int);	// in Filter.cpp
/////////////////////////////////////////////////////////////////////////////
// CMacPacket
void CMacPacket::SetParam(IStream* pStream, DWORD dwSeekPoint)
{
	ATLTRACE(_T("Packet Offset = %d\n"), dwSeekPoint);
	m_pStream = pStream;
	BYTE szBuff[PACKET_HEADER_SIZE];
	ULONG ul;
	LARGE_INTEGER dlib;
	dlib.HighPart = 0;
	dlib.LowPart = dwSeekPoint;
	HRESULT hr = pStream->Seek(dlib, STREAM_SEEK_SET, NULL);
	hr = pStream->Read(szBuff, PACKET_HEADER_SIZE, &ul);
	m_dwSize = *(DWORD*)szBuff - PACKET_HEADER_SIZE;
	CopyMemory(&m_ftime, &szBuff[RCVTIME_OFFSET], sizeof(FILETIME));
	m_dwSeekPoint = PACKET_HEADER_SIZE + dwSeekPoint;
}

void CMacPacket::SetParam(int cb, LPBYTE pb)
{
	m_dwSize = cb - PACKET_HEADER_SIZE;
	CopyMemory(&m_ftime, pb + RCVTIME_OFFSET, sizeof(FILETIME));
	if (m_pbData) delete[] m_pbData;
	m_pbData = new BYTE[cb];
	CopyMemory(m_pbData, pb + PACKET_HEADER_SIZE, cb);
}

HRESULT STDMETHODCALLTYPE CMacPacket::get_Value( 
            /* [retval][out] */ VARIANT __RPC_FAR *pData)
{
	if (!pData) return E_POINTER;
	SAFEARRAYBOUND rgsabound[1];
	rgsabound[0].lLbound = 0;
	rgsabound[0].cElements = m_dwSize;
	SAFEARRAY* pAry = SafeArrayCreate(VT_UI1, 1, rgsabound);
	if (pAry == NULL) return E_OUTOFMEMORY;
	SafeArrayLock(pAry);
	HRESULT hr = GetPacket(NULL, reinterpret_cast<LPBYTE>(pAry->pvData));
	SafeArrayUnlock(pAry);
	if (hr != S_OK) return hr;
	VariantInit(pData);
	pData->vt = VT_ARRAY | VT_UI1;
	pData->parray = pAry;
	return S_OK;
}
        
HRESULT STDMETHODCALLTYPE CMacPacket::get_RecvTime( 
            /* [retval][out] */ DATE __RPC_FAR *pDate)
{
	if (!pDate) return E_POINTER;
	SYSTEMTIME stime;
	FILETIME ftime;
	FileTimeToLocalFileTime(&m_ftime, &ftime);
	FileTimeToSystemTime(&ftime, &stime);
	ATLTRACE(_T("time=%02d:%02d:%02d(%03d)\n"), stime.wHour, stime.wMinute, stime.wSecond, stime.wMilliseconds);
	SystemTimeToVariantTime(&stime, pDate);
	return S_OK;
}

HRESULT STDMETHODCALLTYPE CMacPacket::get_EtherTypeString( 
            /* [retval][out] */ BSTR __RPC_FAR *pResult)
{
	if (!pResult) return E_POINTER;
	HRESULT hr = SetHeader();
	if (hr != S_OK) return hr;
	*pResult = SysAllocString(GetEtherTypeString(m_pbData));
	return S_OK;
}
        
HRESULT STDMETHODCALLTYPE CMacPacket::get_IsMulticast( 
            /* [retval][out] */ VARIANT_BOOL __RPC_FAR *pResult)
{
	if (!pResult) return E_POINTER;
	HRESULT hr = SetHeader();
	if (hr != S_OK) return hr;
	*pResult = IsMulticast(m_pbData) ? VARIANT_TRUE : VARIANT_FALSE;
	return S_OK;
}
        
HRESULT STDMETHODCALLTYPE CMacPacket::get_SrcVendorString( 
            /* [retval][out] */ BSTR __RPC_FAR *pResult)
{
	if (!pResult) return E_POINTER;
	HRESULT hr = SetHeader();
	if (hr != S_OK) return hr;
	*pResult = SysAllocString(GetVendorString(m_pbData + SRC_MAC_ADDRESS));
	return S_OK;
}
        
HRESULT STDMETHODCALLTYPE CMacPacket::get_DestVendorString( 
            /* [retval][out] */ BSTR __RPC_FAR *pResult)
{
	if (!pResult) return E_POINTER;
	HRESULT hr = SetHeader();
	if (hr != S_OK) return hr;
	LPCOLESTR psz = GetMulticastDescription(m_pbData + DST_MAC_ADDRESS, ntohs(*(reinterpret_cast<WORD*>(m_pbData + ETHER_TYPE))));
	if (psz == NULL)
		psz = SysAllocString(GetVendorString(m_pbData + DST_MAC_ADDRESS));
	*pResult = SysAllocString(psz);
	return S_OK;
}
        
HRESULT STDMETHODCALLTYPE CMacPacket::get_FileTime( 
            /* [retval][out] */ CY __RPC_FAR *pResult)
{
	if (!pResult) return E_POINTER;
	CopyMemory(pResult, &m_ftime, sizeof(CY));
	return S_OK;
}
        
HRESULT STDMETHODCALLTYPE CMacPacket::get_InternetProtocolString( 
            /* [retval][out] */ BSTR __RPC_FAR *pResult)
{
	if (!pResult) return E_POINTER;
	HRESULT hr = SetHeader();
	if (hr != S_OK) return hr;
	WORD w = ntohs(*(reinterpret_cast<WORD*>(m_pbData + ETHER_TYPE)));
	if (w == ET_IP)
		*pResult = SysAllocString(GetProtoName(*(m_pbData + IPH_PROTOCOL)));
	else
		*pResult = SysAllocString(L"NOT IP");
	return S_OK;
}
        
HRESULT STDMETHODCALLTYPE CMacPacket::get_SrcMacString( 
            /* [retval][out] */ BSTR __RPC_FAR *pResult)
{
	if (!pResult) return E_POINTER;
	HRESULT hr = SetHeader();
	if (hr != S_OK) return hr;
	OLECHAR szMac[32];
	LPBYTE pb = m_pbData + SRC_MAC_ADDRESS;
	swprintf(szMac, L"%02X-%02X-%02X-%02X-%02X-%02X",
		*pb, *(pb + 1), *(pb + 2), *(pb + 3), *(pb + 4), *(pb + 5));
	*pResult = SysAllocString(szMac);
	return S_OK;
}
        
HRESULT STDMETHODCALLTYPE CMacPacket::get_DestMacString( 
            /* [retval][out] */ BSTR __RPC_FAR *pResult)
{
	if (!pResult) return E_POINTER;
	HRESULT hr = SetHeader();
	if (hr != S_OK) return hr;
	OLECHAR szMac[32];
	LPBYTE pb = m_pbData + DST_MAC_ADDRESS;
	swprintf(szMac, L"%02X-%02X-%02X-%02X-%02X-%02X",
		*pb, *(pb + 1), *(pb + 2), *(pb + 3), *(pb + 4), *(pb + 5));
	*pResult = SysAllocString(szMac);
	return S_OK;
}
        
HRESULT STDMETHODCALLTYPE CMacPacket::get_SrcIPAddressString( 
            /* [retval][out] */ BSTR __RPC_FAR *pResult)
{
	if (!pResult) return E_POINTER;
	HRESULT hr = SetHeader();
	if (hr != S_OK) return hr;
	WORD w = ntohs(*(reinterpret_cast<WORD*>(m_pbData + ETHER_TYPE)));
	if (w == ET_IP)
		GetIPAddress(IPH_SRC, pResult);
	else if (w == ET_ARP || w == ET_RARP)
		GetIPAddress(ARP_SRC_IP, pResult);
	else
		*pResult = SysAllocString(L"No IP Address");
	return S_OK;
}
        
HRESULT STDMETHODCALLTYPE CMacPacket::get_DestIPAddressString( 
            /* [retval][out] */ BSTR __RPC_FAR *pResult)
{
	if (!pResult) return E_POINTER;
	HRESULT hr = SetHeader();
	if (hr != S_OK) return hr;
	WORD w = ntohs(*(reinterpret_cast<WORD*>(m_pbData + ETHER_TYPE)));
	if (w == ET_IP)
		GetIPAddress(IPH_DST, pResult);
	else if (w == ET_ARP || w == ET_RARP)
		GetIPAddress(ARP_DST_IP, pResult);
	else
		*pResult = SysAllocString(L"No IP Address");
	return S_OK;
}

HRESULT STDMETHODCALLTYPE CMacPacket::get_EtherType( 
            /* [retval][out] */ long __RPC_FAR *pResult)
{
	if (!pResult) return E_POINTER;
	HRESULT hr = SetHeader();
	if (hr != S_OK) return hr;
	*pResult = ntohs(*(reinterpret_cast<WORD*>(m_pbData + ETHER_TYPE)));
	return S_OK;
}
        
HRESULT STDMETHODCALLTYPE CMacPacket::get_InternetProtocol(
            /* [retval][out] */ short __RPC_FAR *pResult)
{
	if (!pResult) return E_POINTER;
	HRESULT hr = SetHeader();
	if (hr != S_OK) return hr;
	*pResult = (ntohs(*(reinterpret_cast<WORD*>(m_pbData + ETHER_TYPE))) == ET_IP) ? *(m_pbData + IPH_PROTOCOL) : -1;
	return S_OK;
}
        
HRESULT STDMETHODCALLTYPE CMacPacket::get_Length( 
            /* [retval][out] */ long __RPC_FAR *pResult)
{
	if (!pResult) return E_POINTER;
	*pResult = m_dwSize;
	return S_OK;
}

HRESULT STDMETHODCALLTYPE CMacPacket::get_UnpackedString( 
            /* [retval][out] */ BSTR __RPC_FAR *pResult)
{
	if (!pResult) return E_POINTER;
	LPOLESTR psz = reinterpret_cast<LPOLESTR>(_alloca((m_dwSize * 2 + 1) * sizeof(WCHAR)));
	LPBYTE pb;
	if (m_pStream)
	{
		pb = reinterpret_cast<LPBYTE>(_alloca(m_dwSize));
		HRESULT hr = GetPacket(NULL, pb);
		if (hr != S_OK) return hr;
	}
	else
	{
		pb = m_pbData;
	}
	
	*pResult = SysAllocString(Unpack(psz, pb, m_dwSize));
	return S_OK;
}

HRESULT STDMETHODCALLTYPE CMacPacket::get_AsAnsiString( 
            /* [in] */ long offset,
            /* [in] */ VARIANT length,
            /* [retval][out] */ BSTR __RPC_FAR *pResult)
{
	if (!pResult) return E_POINTER;
	if (offset >= m_dwSize)
	{
		*pResult = SysAllocString(L"");
		return S_OK;
	}
	USES_CONVERSION;
	LPBYTE pb;
	if (m_pStream)
	{
		pb = reinterpret_cast<LPBYTE>(_alloca(m_dwSize));
		HRESULT hr = GetPacket(NULL, pb);
		if (hr != S_OK) return hr;
	}
	else
	{
		pb = m_pbData;
	}
	int cb;
	HRESULT hr = GetLength(length, offset, cb);
	if (hr != S_OK) return hr;
	LPWSTR pw = reinterpret_cast<LPWSTR>(_alloca(cb * 2 + 8));
	int ix = MultiByteToWideChar(GetACP(), 0, reinterpret_cast<LPSTR>(pb + offset), cb, pw, cb * 2 + 8);
	*(pw + ix) = L'\0';
	*pResult = SysAllocString(pw);
	return S_OK;
}

HRESULT STDMETHODCALLTYPE CMacPacket::get_AsString( 
            /* [in] */ long offset,
            /* [in] */ VARIANT length,
            /* [retval][out] */ BSTR __RPC_FAR *pResult)
{
	if (!pResult) return E_POINTER;
	if (offset >= m_dwSize)
	{
		*pResult = SysAllocString(L"");
		return S_OK;
	}
	LPBYTE pb;
	if (m_pStream)
	{
		pb = reinterpret_cast<LPBYTE>(_alloca(m_dwSize));
		HRESULT hr = GetPacket(NULL, pb);
		if (hr != S_OK) return hr;
	}
	else
	{
		pb = m_pbData;
	}
	int cb;
	HRESULT hr = GetLength(length, offset, cb);
	if (hr != S_OK) return hr;
	*pResult = SysAllocStringByteLen(reinterpret_cast<LPSTR>(pb + offset), cb);
	return S_OK;
}

HRESULT CMacPacket::GetLength(VARIANT vsrc, int offset, int& cb)
{
	if (vsrc.vt == VT_ERROR)
	{
		cb = m_dwSize - offset;
		return S_OK;
	}
	VARIANT v;
	VariantInit(&v);
	HRESULT hr = VariantChangeType(&v, &vsrc, 0, VT_I4);
	if (hr != S_OK) return hr;
	cb = v.lVal;

	if ((cb + offset) > m_dwSize)
	{
		cb = m_dwSize - offset;
	}
	return S_OK;
}

HRESULT STDMETHODCALLTYPE CMacPacket::GetPacket( 
            /* [out] */ long __RPC_FAR *pLength,
            /* [out][in] */ unsigned char __RPC_FAR buff[ 1514 ])
{
	if (!buff) return E_POINTER;
	if (m_pStream)
	{
		LARGE_INTEGER dlib;
		dlib.HighPart = 0;
		dlib.LowPart = m_dwSeekPoint;
		m_pStream->Seek(dlib, STREAM_SEEK_SET, NULL);
		ULONG len;
		HRESULT hr = m_pStream->Read(buff, m_dwSize, &len);
		if (m_pbData == NULL && hr == S_OK)
		{
			m_pbData = new BYTE[ARP_MAX];
			CopyMemory(m_pbData, buff, ARP_MAX);
		}
		if (hr != S_OK) return hr;
		if (pLength) *pLength = len;
	}
	else
	{
		CopyMemory(buff, m_pbData, m_dwSize);
		if (pLength)
			*pLength = m_dwSize;
	}
	return S_OK;
}
        
HRESULT STDMETHODCALLTYPE CMacPacket::GetRecvTime( 
            /* [out] */ FILETIME __RPC_FAR *pFileTime)
{
	if (!pFileTime) return E_POINTER;
	CopyMemory(pFileTime, &m_ftime, sizeof(FILETIME));
	return S_OK;
}
        
HRESULT STDMETHODCALLTYPE CMacPacket::GetLength( 
            /* [out] */ long __RPC_FAR *pLength)
{
	if (!pLength) return E_POINTER;
	*pLength = m_dwSize;
	return S_OK;
}

///////////////////////////////////////////////////////////////////
STDMETHODIMP CMacPacket::InterfaceSupportsErrorInfo(REFIID riid)
{
	static const IID* arr[] = 
	{
		&IID_IMacPacket
	};
	for (int i=0; i < sizeof(arr) / sizeof(arr[0]); i++)
	{
		if (::InlineIsEqualGUID(*arr[i],riid))
			return S_OK;
	}
	return S_FALSE;
}

HRESULT CMacPacket::SetHeader()
{
	if (m_pbData) return S_OK;
	m_pbData = new BYTE[ARP_MAX];
	if (!m_pbData) return E_OUTOFMEMORY;
	ULONG ul;
	LARGE_INTEGER dlib;
	dlib.HighPart = 0;
	dlib.LowPart = m_dwSeekPoint;
	m_pStream->Seek(dlib, STREAM_SEEK_SET, NULL);
	HRESULT hr = m_pStream->Read(m_pbData, ARP_MAX, &ul);
	return hr;
}

void CMacPacket::GetIPAddress(int offset, BSTR* pResult)
{
	OLECHAR sz[24];
	swprintf(sz, L"%d.%d.%d.%d", *(m_pbData + offset), *(m_pbData + offset + 1),
		*(m_pbData + offset + 2), *(m_pbData + offset + 3));
	*pResult = SysAllocString(sz);
}

