/*
 NCapDD.c - NDIS Capture kernel-mode driver

 Free software Copyright (c) 1998-2000 arton

 Log:
   13 Aug. 2002 - Version 3.0

   21 Dec. 2000 - Version 2.3
         Changed 1000000 -> 1000 for som bad Frequency
   22 Jan. 2000 - Version 2.2
         Changed CapReceiveIndicate() to check IRQL when copy the timestamp from stack to UserMem.
   11 Apr. 1999 - Version 2.1
         Changed to bind by App Lib. (NCapDD can't enum Services key)
   28 Mar. 1999 - Version 2.0
         Add dynamic binding - suggested by Lewis Perin (thanks!)
   26 Dec. 1998 - Version 1.0
         Based on DDK's packet driver sample
*/

#include <ntddk.h>
#include <ntiologc.h>
#include <ndis.h>
#include <ntddndis.h>

#include "NCap.h"
#include "NCapDD.h"


NTSTATUS DriverEntry(IN PDRIVER_OBJECT pDriverObject,
                     IN PUNICODE_STRING pRegistryPath);
#pragma NDIS_INIT_FUNCTION(DriverEntry)

static NTSTATUS ReadRegistry(IN PWSTR*, IN PWSTR*, IN PUNICODE_STRING);
#pragma NDIS_INIT_FUNCTION(ReadRegistry)

static NTSTATUS CapQueryRegistryRoutine(
    IN PWSTR     ValueName,
    IN ULONG     ValueType,
    IN PVOID     ValueData,
    IN ULONG     ValueLength,
    IN PVOID     Context,
    IN PVOID     EntryContext);
#pragma NDIS_INIT_FUNCTION(CapQueryRegistryRoutine)

NDIS_STRING ProtoName = NDIS_STRING_CONST("CaptureDriver");
static UINT SPacketFilter = NDIS_PACKET_TYPE_PROMISCUOUS;

static LONGLONG llStartTick = 0;     // Hz
static LONGLONG llTickDivisor = 0;   // Divisor for Adjusting to 100ns
static LONGLONG llStartSysTime = 0;  // first digit = 100ns

/*
 convenience (routines) store
*/
NTSTATUS finishIrp(IN OUT PIRP airpP, IN NTSTATUS astatus, IN ULONG aninfo)
{
  airpP->IoStatus.Status = astatus;
  airpP->IoStatus.Information = aninfo;
  IoCompleteRequest(airpP, IO_NO_INCREMENT);
  return astatus;
}

void EmptyOpenInstance(POPEN_INSTANCE Open)
{
    NdisFreeSpinLock(Open->RequestSpinLock);
    if (Open->PacketPool)
    {
        NdisFreePacketPool(Open->PacketPool);
	Open->PacketPool = NULL;
    }
    ExFreePool(Open);
}

/*
 from http://www.panix.com/~perin/packetbugs.html 
*/
void CapCancelReadIrp(IN PDEVICE_OBJECT adeviceObjectP, IN PIRP airpP)
{
  PIO_STACK_LOCATION ioslP = IoGetCurrentIrpStackLocation(airpP);
  POPEN_INSTANCE oiP = ioslP->FileObject->FsContext;
  BOOLEAN foundIt = FALSE;
  PLIST_ENTRY nextLinkP;

  for (nextLinkP = oiP->RcvList.Flink;
       nextLinkP != &oiP->RcvList;
       nextLinkP = nextLinkP->Flink) {
    PCapReserved reservedP =
      CONTAINING_RECORD(nextLinkP, CapReserved, ListElement);
    PNDIS_PACKET packetP  =
      CONTAINING_RECORD(reservedP, NDIS_PACKET, ProtocolReserved);
    if (reservedP->Irp == airpP) {
      foundIt = TRUE;
      RemoveEntryList(nextLinkP);
      NdisReinitializePacket(packetP); // recycle the packet
      NdisFreePacket(packetP);  // Put the packet on the free queue
      break;
    }
  }

  IoReleaseCancelSpinLock(airpP->CancelIrql);
  if (foundIt) finishIrp(airpP, STATUS_CANCELLED, 0);
}

/*
 Required Functions (But useless in this program)
*/
VOID
CapStatus(
    IN NDIS_HANDLE   ProtocolBindingContext,
    IN NDIS_STATUS   Status,
    IN PVOID         StatusBuffer,
    IN UINT          StatusBufferSize
    )
{

    return;
}


VOID
CapStatusComplete(
    IN NDIS_HANDLE  ProtocolBindingContext
    )
{
    return;
}

VOID
CapResetComplete(
    IN NDIS_HANDLE  ProtocolBindingContext,
    IN NDIS_STATUS  Status
    )
{
}

VOID CapReceiveComplete(IN NDIS_HANDLE abindingContext)
{
}

VOID
CapSendComplete(
    IN NDIS_HANDLE   ProtocolBindingContext,
    IN PNDIS_PACKET  pPacket,
    IN NDIS_STATUS   Status
    )
{
    PIRP Irp=RESERVED(pPacket)->Irp;

    //
    //  recyle the packet
    //
    NdisReinitializePacket(pPacket);

    //
    //  Put the packet back on the free list
    //
    if (KeGetCurrentIrql() == DISPATCH_LEVEL)
    {
        NdisDprFreePacket(pPacket);
    }
    else
    {
        NdisFreePacket(pPacket);
    }

    finishIrp(Irp, Status, 0);
}

/*
 NDIS Request Complete Dispatch
*/
VOID
CapRequestComplete(
    IN NDIS_HANDLE   ProtocolBindingContext,
    IN PNDIS_REQUEST NdisRequest,
    IN NDIS_STATUS   Status
    )
{
    POPEN_INSTANCE Open= (POPEN_INSTANCE)ProtocolBindingContext;

    PCapRequest pRequest=CONTAINING_RECORD(NdisRequest,CapRequest,Request);
    PIRP Irp=pRequest->Irp;
    NdisInterlockedInsertTailList(&Open->RequestList, 
				  &pRequest->ListElement,
				  &Open->RequestSpinLock);
    if (Irp)
    {
        finishIrp(Irp, Status, 0);
	if (Irp == Open->OpenCloseIrp)
	{
	    if (NT_SUCCESS(Status))
	    {
		Open->OpenCloseIrp = NULL;
	    }
	}
    }
}

/*
 I/O Control
*/
NTSTATUS CapIoControl(IN PDEVICE_OBJECT pDeviceObject, IN PIRP Irp)
{
    PLIST_ENTRY         RequestListEntry;
    PCapRequest   pRequest;

    PIO_STACK_LOCATION IrpSp = IoGetCurrentIrpStackLocation(Irp);
    POPEN_INSTANCE Open=IrpSp->FileObject->FsContext;
    NTSTATUS Status = STATUS_SUCCESS;
    UINT lengthNeeded;            // how much buffer space we need to write to
    UINT lengthTransferred = 0;   // how much buffer space we did write to

    switch (IrpSp->Parameters.DeviceIoControl.IoControlCode)
    {
    case DIOC_CAPGetMacAddress:
        lengthNeeded = MacAddressSize;
	if ((Irp->AssociatedIrp.SystemBuffer == NULL) ||
	    (IrpSp->Parameters.DeviceIoControl.OutputBufferLength < lengthNeeded))
	    Status = STATUS_INVALID_PARAMETER;
	else
	{
	    RtlCopyBytes(Irp->AssociatedIrp.SystemBuffer, Open->MacAddress, MacAddressSize);
	    lengthTransferred = lengthNeeded;
	}
	return finishIrp(Irp, Status, lengthTransferred);
    case DIOC_CAPGetTick:
        lengthNeeded = 16;
        if ((Irp->AssociatedIrp.SystemBuffer == NULL) ||
            (IrpSp->Parameters.DeviceIoControl.OutputBufferLength < lengthNeeded))
            Status = STATUS_INVALID_PARAMETER;
        else
        {
            LARGE_INTEGER tick[2];
            tick[0] = KeQueryPerformanceCounter(&tick[1]);
            RtlCopyBytes(Irp->AssociatedIrp.SystemBuffer, tick, 16);
            lengthTransferred = lengthNeeded;
        }
        return finishIrp(Irp, Status, lengthTransferred);
    default:
        return finishIrp(Irp, NDIS_STATUS_FAILURE, 0);
    }
}

/*
 Open
*/
VOID CapOpenAdapterComplete(
    IN NDIS_HANDLE  ProtocolBindingContext,
    IN NDIS_STATUS  Status,
    IN NDIS_STATUS  OpenErrorStatus
    )
{

    PIRP              Irp;
    POPEN_INSTANCE    Open;

    Open= (POPEN_INSTANCE)ProtocolBindingContext;

    //
    //  get the open irp
    //
    Irp=Open->OpenCloseIrp;

    if (Status == NDIS_STATUS_SUCCESS)
    {
        PCapRequest MacReq = (PCapRequest)
	  NdisInterlockedRemoveHeadList(&Open->RequestList, &Open->RequestSpinLock);
        PCapRequest req = (PCapRequest)
	  NdisInterlockedRemoveHeadList(&Open->RequestList, &Open->RequestSpinLock);
	if (MacReq && req)
	{
	    PNDIS_REQUEST pReq;
	    MacReq->Irp = NULL;
	    pReq = &MacReq->Request;
	    pReq->RequestType = NdisRequestQueryInformation;
	    pReq->DATA.QUERY_INFORMATION.Oid = OID_802_3_CURRENT_ADDRESS;
	    pReq->DATA.QUERY_INFORMATION.InformationBuffer = Open->MacAddress;
	    pReq->DATA.QUERY_INFORMATION.InformationBufferLength = MacAddressSize;
	    NdisRequest(&Open->MacAddressReqStatus, Open->AdapterHandle, pReq);
	    if (Open->MacAddressReqStatus != NDIS_STATUS_PENDING)
	    {
	        CapRequestComplete((NDIS_HANDLE)Open, pReq, Open->MacAddressReqStatus);
	    }
	    req->Irp = Irp;
	    pReq = &req->Request;
	    pReq->RequestType = NdisRequestSetInformation;
	    pReq->DATA.SET_INFORMATION.Oid = OID_GEN_CURRENT_PACKET_FILTER;
	    pReq->DATA.SET_INFORMATION.InformationBuffer = &SPacketFilter;
	    pReq->DATA.SET_INFORMATION.InformationBufferLength = sizeof(SPacketFilter);
	    NdisRequest(&Open->SetFilterReqStatus, Open->AdapterHandle, pReq);
	    if (Open->SetFilterReqStatus != NDIS_STATUS_PENDING)
	    {
	        CapRequestComplete((NDIS_HANDLE)Open, pReq, Open->SetFilterReqStatus);
	    }
	}
	else
	{
	    Status = STATUS_INSUFFICIENT_RESOURCES;
	}
    }

    if (Status != NDIS_STATUS_SUCCESS)
    {
        finishIrp(Irp, Status, 0);
	EmptyOpenInstance(Open);
    }
}

VOID CapTransferDataComplete(
    IN NDIS_HANDLE   ProtocolBindingContext,
    IN PNDIS_PACKET  pPacket,
    IN NDIS_STATUS   Status,
    IN UINT          BytesTransfered
    )
{
    POPEN_INSTANCE Open= (POPEN_INSTANCE)ProtocolBindingContext;

    PIRP Irp=RESERVED(pPacket)->Irp;

    PIO_STACK_LOCATION IrpSp = IoGetCurrentIrpStackLocation(Irp);

    PMDL pMdl=RESERVED(pPacket)->pMdl;

    //
    //  Free the MDL that we allocated
    //
    IoFreeMdl(pMdl);

    //
    //  recylcle the packet
    //
    NdisReinitializePacket(pPacket);

    //
    //  Put the packet on the free queue
    //
    if (KeGetCurrentIrql() == DISPATCH_LEVEL)
    {
        NdisDprFreePacket(pPacket);
    }
    else
    {
        NdisFreePacket(pPacket);
    }

    if (Status != STATUS_CANCELLED)
      finishIrp(Irp, Status, BytesTransfered + sizeof(CapTimestamp) + ETHERNET_HEADER_LENGTH);
}

static UINT            Medium;
static NDIS_MEDIUM     MediumArray=NdisMedium802_3;
NTSTATUS CapOpen(IN PDEVICE_OBJECT DeviceObject, IN PIRP Irp)
{
    PDEVICE_EXTENSION DeviceExtension;

    POPEN_INSTANCE    Open;

    PIO_STACK_LOCATION  IrpSp;

    NDIS_STATUS     Status;
    NDIS_STATUS     ErrorStatus;

    UINT            i;

    DeviceExtension = DeviceObject->DeviceExtension;

    IrpSp = IoGetCurrentIrpStackLocation(Irp);

    //
    //  allocate some memory for the open structure
    //
    Open=ExAllocatePool(NonPagedPool,sizeof(OPEN_INSTANCE));

    if (Open==NULL) {
        //
        // no memory
        //
        return finishIrp(Irp, STATUS_INSUFFICIENT_RESOURCES, 0);
    }

    RtlZeroMemory(
        Open,
        sizeof(OPEN_INSTANCE)
        );

    //
    //  Save or open here
    //
    IrpSp->FileObject->FsContext=Open;

    Open->DeviceExtension=DeviceExtension;

    //
    //  Save the Irp here for the completeion routine to retrieve
    //
    Open->OpenCloseIrp=Irp;

    //
    //  Allocate a packet pool for our xmit and receive packets
    //
    NdisAllocatePacketPool(
        &Status,
        &Open->PacketPool,
        TRANSMIT_PACKETS,
        sizeof(CapReserved));

    if (Status != NDIS_STATUS_SUCCESS) {

        ExFreePool(Open);
        return finishIrp(Irp, STATUS_INSUFFICIENT_RESOURCES, 0);
    }

    //
    //  list to hold irp's want to reset the adapter
    //
    InitializeListHead(&Open->ResetIrpList);


    //
    //  Initialize list for holding pending read requests
    //
    KeInitializeSpinLock(&Open->RcvQSpinLock);
    InitializeListHead(&Open->RcvList);

    //
    //  Initialize the request list
    //
    NdisAllocateSpinLock(&Open->RequestSpinLock);
    InitializeListHead(&Open->RequestList);


    //
    //  link up the request stored in our open block
    //
    for (i=0;i<MAX_REQUESTS;i++) {
        NdisInterlockedInsertTailList(
            &Open->RequestList,
            &Open->Requests[i].ListElement,
            &Open->RequestSpinLock);

    }


    IoMarkIrpPending(Irp);
    Irp->IoStatus.Status = STATUS_PENDING;

    //
    //  Try to open the MAC
    //
    NdisOpenAdapter(
        &Status,
        &ErrorStatus,
        &Open->AdapterHandle,
        &Medium,
        &MediumArray,
        1,
        DeviceExtension->NdisProtocolHandle,
        Open,
        &DeviceExtension->AdapterName,
        0,
        NULL);


    if (Status != NDIS_STATUS_PENDING) {

        CapOpenAdapterComplete(
            Open,
            Status,
            NDIS_STATUS_SUCCESS
            );


    }



    return(STATUS_PENDING);
}


/*
 Read
*/
NTSTATUS CapRead(IN PDEVICE_OBJECT DeviceObject, IN PIRP Irp)
{
    POPEN_INSTANCE      Open;
    PIO_STACK_LOCATION  IrpSp;
    PNDIS_PACKET        pPacket;
    PMDL                pMdl;
    NDIS_STATUS         Status;
    KIRQL irql;

    IrpSp = IoGetCurrentIrpStackLocation(Irp);

    Open=IrpSp->FileObject->FsContext;

    //
    //  See if the buffer is atleast big enough to hold the
    //  ethernet header
    //
    if (IrpSp->Parameters.Read.Length < (ETHERNET_HEADER_LENGTH + sizeof(CapTimestamp))) {

        return finishIrp(Irp, STATUS_UNSUCCESSFUL, 0);
    }

    //
    //  Allocate an MDL to map the portion of the buffer following the
    //  header
    //
    pMdl=IoAllocateMdl(
              MmGetMdlVirtualAddress(Irp->MdlAddress),
              MmGetMdlByteCount(Irp->MdlAddress),
              FALSE,
              FALSE,
              NULL
              );


    if (pMdl == NULL) {
        return finishIrp(Irp, STATUS_UNSUCCESSFUL, 0);
    }

    //
    //  Build the mdl to point to the the portion of the buffer followin
    //  the header
    //
    IoBuildPartialMdl(
        Irp->MdlAddress,
        pMdl,
        ((PUCHAR)MmGetMdlVirtualAddress(Irp->MdlAddress))+ETHERNET_HEADER_LENGTH + sizeof(CapTimestamp),
        0
        );

    //
    //  Clear the next link in the new MDL
    //
    pMdl->Next=NULL;

    //
    //  Try to get a packet from our list of free ones
    //

    NdisAllocatePacket(
        &Status,
        &pPacket,
        Open->PacketPool
        );

    if (Status != NDIS_STATUS_SUCCESS) {

        IoFreeMdl(pMdl);
	return finishIrp(Irp, STATUS_UNSUCCESSFUL, 0);
    }

    //
    //  Get a pointer to the packet itself
    //

    RESERVED(pPacket)->Irp=Irp;
    RESERVED(pPacket)->pMdl=pMdl;

    IoMarkIrpPending(Irp);
    Irp->IoStatus.Status = STATUS_PENDING;

    //
    //  Attach our new MDL to the packet
    //
    NdisChainBufferAtFront(pPacket,pMdl);
    ExInterlockedInsertTailList(
        &Open->RcvList,
        &RESERVED(pPacket)->ListElement,
        &Open->RcvQSpinLock);

    IoAcquireCancelSpinLock(&irql);
    IoSetCancelRoutine(Irp, CapCancelReadIrp);
    IoReleaseCancelSpinLock(irql);

    return(STATUS_PENDING);
}

NDIS_STATUS CapReceiveIndicate(
    IN NDIS_HANDLE ProtocolBindingContext,
    IN NDIS_HANDLE MacReceiveContext,
    IN PVOID       HeaderBuffer,
    IN UINT        HeaderBufferSize,
    IN PVOID       LookAheadBuffer,
    IN UINT        LookaheadBufferSize,
    IN UINT        PacketSize
   )
{
    POPEN_INSTANCE      Open;
    PIO_STACK_LOCATION  IrpSp;
    PIRP                Irp;
    PLIST_ENTRY         PacketListEntry;
    PNDIS_PACKET        pPacket;
    ULONG               SizeToTransfer;
    NDIS_STATUS         Status;
    UINT                BytesTransfered;
    ULONG               BufferLength;
    CapTimestamp        timestamp;
    LONGLONG llFreq;
    UCHAR	        puDummy[16];

    PCapReserved    Reserved;

    if (HeaderBufferSize != ETHERNET_HEADER_LENGTH) {

        return NDIS_STATUS_NOT_ACCEPTED;
    }

    Open= (POPEN_INSTANCE)ProtocolBindingContext;

    //
    //  See if there are any pending read that we can satisfy
    //
    
    PacketListEntry=ExInterlockedRemoveHeadList(
                        &Open->RcvList,
                        &Open->RcvQSpinLock
                        );

    if (PacketListEntry == NULL) {

        return NDIS_STATUS_NOT_ACCEPTED;
    }

    Reserved=CONTAINING_RECORD(PacketListEntry,CapReserved,ListElement);
    pPacket=CONTAINING_RECORD(Reserved,NDIS_PACKET,ProtocolReserved);

    timestamp.Reserved = 0;
    timestamp.timeRecved = (KeQueryPerformanceCounter(&llFreq)).QuadPart - llStartTick;
    timestamp.timeRecved *= 1000;
    timestamp.timeRecved /= llTickDivisor;
    timestamp.timeRecved += llStartSysTime;

    Irp=RESERVED(pPacket)->Irp;
    IrpSp = IoGetCurrentIrpStackLocation(Irp);

    //
    //  This is the length of our partial MDL
    //
    BufferLength=IrpSp->Parameters.Read.Length-(ETHERNET_HEADER_LENGTH + sizeof(CapTimestamp));

    //
    //  Find out how much to transfer
    //
    SizeToTransfer = (PacketSize < BufferLength) ?
                       PacketSize : BufferLength;

    //
    //  copy the ethernet header into the actual readbuffer
    //
    NdisMoveToMappedMemory(
	    MmGetSystemAddressForMdl(Irp->MdlAddress),
	    &timestamp,
	    sizeof(timestamp));
    NdisMoveMappedMemory(
    	((PUCHAR)MmGetSystemAddressForMdl(Irp->MdlAddress)) + sizeof(timestamp),
        HeaderBuffer,
        HeaderBufferSize);

    //
    //  Call the Mac to transfer the packet
    //
    NdisTransferData(
        &Status,
        Open->AdapterHandle,
        MacReceiveContext,
        0,
        SizeToTransfer,
        pPacket,
        &BytesTransfered);

    if (Status != NDIS_STATUS_PENDING) {

        //
        //  If it didn't pend, call the completeion routine now
        //
        CapTransferDataComplete(
            Open,
            pPacket,
            Status,
            BytesTransfered
            );
    }

    return NDIS_STATUS_SUCCESS;
}


/*
 Close
*/

VOID CapCloseAdapterComplete(
    IN NDIS_HANDLE  ProtocolBindingContext,
    IN NDIS_STATUS  Status
    )
{
    POPEN_INSTANCE    Open;
    PIRP              Irp;

    Open= (POPEN_INSTANCE)ProtocolBindingContext;

    Irp=Open->OpenCloseIrp;

    EmptyOpenInstance(Open);

    Irp->IoStatus.Status = STATUS_SUCCESS;
    IoCompleteRequest(Irp, IO_NO_INCREMENT);
}

NTSTATUS CapClose(IN PDEVICE_OBJECT DeviceObject, IN PIRP Irp)
{

    POPEN_INSTANCE    Open;
    NDIS_STATUS     Status;
    PIO_STACK_LOCATION  IrpSp;

    IrpSp = IoGetCurrentIrpStackLocation(Irp);

    Open=IrpSp->FileObject->FsContext;

    //
    //  Save the IRP
    //
    Open->OpenCloseIrp=Irp;

    IoMarkIrpPending(Irp);
    Irp->IoStatus.Status = STATUS_PENDING;

    //
    //  close the adapter
    //
    NdisCloseAdapter(
        &Status,
        Open->AdapterHandle
        );


    if (Status != NDIS_STATUS_PENDING) {

        CapCloseAdapterComplete(
            Open,
            Status
            );


    }

    return(STATUS_PENDING);
}


/*
 Cleanup
*/
NTSTATUS CapCleanup(IN PDEVICE_OBJECT adeviceObjectP, IN PIRP airpP)
{
  PIO_STACK_LOCATION ioslP = IoGetCurrentIrpStackLocation(airpP);
  POPEN_INSTANCE oiP = ioslP->FileObject->FsContext;
  PLIST_ENTRY listEntryP;

  IoMarkIrpPending(airpP);
  airpP->IoStatus.Status = STATUS_PENDING;

  do {
    KIRQL irql;
    IoAcquireCancelSpinLock(&irql);
    listEntryP = (IsListEmpty(&oiP->RcvList)) ? NULL :
      RemoveHeadList(&oiP->RcvList);
    IoReleaseCancelSpinLock(irql);
    if (listEntryP != NULL) {
      PCapReserved reservedP =
        CONTAINING_RECORD(listEntryP, CapReserved, ListElement);
      PNDIS_PACKET packetP =
        CONTAINING_RECORD(reservedP, NDIS_PACKET, ProtocolReserved);
      CapTransferDataComplete(oiP, packetP, STATUS_CANCELLED, 0);
    }
  } while (listEntryP != NULL);
  return finishIrp(airpP, STATUS_SUCCESS, 0);
}

VOID CapUnload(IN PDRIVER_OBJECT DriverObject)
{
    PDEVICE_OBJECT     DeviceObject;
    PDEVICE_OBJECT     OldDeviceObject;
    PDEVICE_EXTENSION  DeviceExtension;

    NDIS_HANDLE        NdisProtocolHandle;
    NDIS_STATUS        Status;

    DeviceObject    = DriverObject->DeviceObject;
    while (DeviceObject != NULL) {

        DeviceExtension = DeviceObject->DeviceExtension;
        NdisProtocolHandle=DeviceExtension->NdisProtocolHandle;

        if (DeviceExtension->BindString != NULL) {
            ExFreePool(DeviceExtension->BindString);
        }

        if (DeviceExtension->ExportString != NULL) {
            ExFreePool(DeviceExtension->ExportString);
        }

        OldDeviceObject=DeviceObject;
        DeviceObject=DeviceObject->NextDevice;

        IoDeleteDevice(OldDeviceObject);
    }
    NdisDeregisterProtocol(
        &Status,
        NdisProtocolHandle
        );
}

/*
 Main entry
*/
NTSTATUS DriverEntry(IN PDRIVER_OBJECT pDriverObject,
                     IN PUNICODE_STRING pRegistryPath)
{
    NTSTATUS Status = NDIS_STATUS_FAILURE;
    ULONG          DevicesCreated=0;
    NDIS_PROTOCOL_CHARACTERISTICS  ProtocolChar;

    UNICODE_STRING MacDriverName;
    UNICODE_STRING UnicodeDeviceName;

    PWSTR          BindString;
    PWSTR          ExportString;
    PWSTR          BindStringSave;
    PWSTR          ExportStringSave;

    PDEVICE_OBJECT pDeviceObject = NULL;
    PDEVICE_EXTENSION pDeviceExtension = NULL;

    NDIS_HANDLE    NdisProtocolHandle;

    DebugPrint(("DriverEntry\n"));

    RtlZeroMemory(&ProtocolChar, sizeof(NDIS_PROTOCOL_CHARACTERISTICS));
    ProtocolChar.MajorNdisVersion            = 3;
    ProtocolChar.MinorNdisVersion            = 0;
    ProtocolChar.Reserved                    = 0;
    ProtocolChar.OpenAdapterCompleteHandler  = CapOpenAdapterComplete;
    ProtocolChar.CloseAdapterCompleteHandler = CapCloseAdapterComplete;
    ProtocolChar.SendCompleteHandler         = CapSendComplete;
    ProtocolChar.TransferDataCompleteHandler = CapTransferDataComplete;
    ProtocolChar.ResetCompleteHandler        = CapResetComplete;
    ProtocolChar.RequestCompleteHandler      = CapRequestComplete;
    ProtocolChar.ReceiveHandler              = CapReceiveIndicate;
    ProtocolChar.ReceiveCompleteHandler      = CapReceiveComplete;
    ProtocolChar.StatusHandler               = CapStatus;
    ProtocolChar.StatusCompleteHandler       = CapStatusComplete;
    ProtocolChar.Name                        = ProtoName;

    NdisRegisterProtocol(
        &Status,
        &NdisProtocolHandle,
        &ProtocolChar,
        sizeof(NDIS_PROTOCOL_CHARACTERISTICS));

    if (Status != NDIS_STATUS_SUCCESS) 
    {
        DebugPrint(("NdisRegisterProtocol (%d)\n", Status));
        return Status;
    }

    pDriverObject->MajorFunction[IRP_MJ_CREATE] = CapOpen;
    pDriverObject->MajorFunction[IRP_MJ_CLOSE] = CapClose;
    pDriverObject->MajorFunction[IRP_MJ_READ] = CapRead;
    pDriverObject->MajorFunction[IRP_MJ_CLEANUP] = CapCleanup;
    pDriverObject->MajorFunction[IRP_MJ_DEVICE_CONTROL] = CapIoControl;
    pDriverObject->DriverUnload = CapUnload;

    Status = ReadRegistry(&BindString, &ExportString, pRegistryPath);
    DebugPrint(("ReadRegistry %d\n", Status));
    if (Status == STATUS_SUCCESS)
    {
        BindStringSave = BindString;
	ExportStringSave = ExportString;
    
	while (*BindString && *ExportString)
	{
	    RtlInitUnicodeString(&MacDriverName, BindString);
	    RtlInitUnicodeString(&UnicodeDeviceName, ExportString);
	    BindString   += (MacDriverName.Length+sizeof(UNICODE_NULL))/sizeof(WCHAR);

	    ExportString += (UnicodeDeviceName.Length+sizeof(UNICODE_NULL))/sizeof(WCHAR);

	    Status = IoCreateDevice(pDriverObject, sizeof(DEVICE_EXTENSION),
				    &UnicodeDeviceName, FILE_DEVICE_PROTOCOL,
				    0, FALSE, &pDeviceObject);
	    if (Status != STATUS_SUCCESS)
	        break;

	    DevicesCreated++;
	    pDeviceObject->Flags |= DO_DIRECT_IO;
	    pDeviceExtension = (PDEVICE_EXTENSION) pDeviceObject->DeviceExtension;
	    pDeviceExtension->DeviceObject = pDeviceObject;
	    pDeviceExtension->AdapterName = MacDriverName;
	    if (DevicesCreated == 1)
	    {
	        pDeviceExtension->BindString = BindStringSave;
		pDeviceExtension->ExportString = ExportStringSave;
	    }
	    pDeviceExtension->NdisProtocolHandle = NdisProtocolHandle;
	}
	if (DevicesCreated > 0)
	{
	  //
	  //  Initialize Timer Tick
	  //
	  LONGLONG llFreq;
	  llStartTick = (KeQueryPerformanceCounter((PLARGE_INTEGER)&llFreq)).QuadPart;
	  NdisGetCurrentSystemTime(&llStartSysTime);
	  // get divisor for 100ns
//	  llTickDivisor = llFreq / 10000000;
	  llTickDivisor = llFreq / 10000;    // 0.1ns
	  if (llTickDivisor == 0) llTickDivisor = 1;  // Who use that CPU (<10Mhz) ?
	  // cut invalid digits (10Ms, 1Ms, 100ns)
	  llStartSysTime /= 1000;
	  llStartSysTime *= 1000;

	  return STATUS_SUCCESS;
	}
	ExFreePool(BindStringSave);
	ExFreePool(ExportStringSave);
    }
    NdisDeregisterProtocol(
        &Status,
        NdisProtocolHandle
        );

    return STATUS_UNSUCCESSFUL;
}
  

NTSTATUS
ReadRegistry(
    IN  PWSTR              *MacDriverName,
    IN  PWSTR              *DriverName,
    IN  PUNICODE_STRING     RegistryPath
    )

{

    NTSTATUS   Status;

    RTL_QUERY_REGISTRY_TABLE ParamTable[4];

    PWSTR      Bind       = L"Bind";
    PWSTR      Export     = L"Export";
    PWSTR      Linkage    = L"Linkage";

    PWCHAR     Path;

    Path=ExAllocatePool(
             PagedPool,
             RegistryPath->Length+sizeof(WCHAR)
             );

    if (Path == NULL) {
        return STATUS_INSUFFICIENT_RESOURCES;
    }

    RtlZeroMemory(
        Path,
        RegistryPath->Length+sizeof(WCHAR)
        );

    RtlCopyMemory(
        Path,
        RegistryPath->Buffer,
        RegistryPath->Length
        );

    RtlZeroMemory(
        ParamTable,
        sizeof(ParamTable)
        );

    //
    //  change to the linkage key
    //

    ParamTable[0].QueryRoutine = NULL;
    ParamTable[0].Flags = RTL_QUERY_REGISTRY_SUBKEY;
    ParamTable[0].Name = Linkage;


    //
    //  Get the name of the mac driver we should bind to
    //

    ParamTable[1].QueryRoutine = CapQueryRegistryRoutine;
    ParamTable[1].Flags = RTL_QUERY_REGISTRY_REQUIRED |
                          RTL_QUERY_REGISTRY_NOEXPAND;

    ParamTable[1].Name = Bind;
    ParamTable[1].EntryContext = (PVOID)MacDriverName;
    ParamTable[1].DefaultType = REG_MULTI_SZ;

    //
    //  Get the name that we should use for the driver object
    //

    ParamTable[2].QueryRoutine = CapQueryRegistryRoutine;
    ParamTable[2].Flags = RTL_QUERY_REGISTRY_REQUIRED |
                          RTL_QUERY_REGISTRY_NOEXPAND;

    ParamTable[2].Name = Export;
    ParamTable[2].EntryContext = (PVOID)DriverName;
    ParamTable[2].DefaultType = REG_MULTI_SZ;


    Status=RtlQueryRegistryValues(
               RTL_REGISTRY_ABSOLUTE,
               Path,
               ParamTable,
               NULL,
               NULL
               );


    ExFreePool(Path);

    return Status;
}

NTSTATUS CapQueryRegistryRoutine(
    IN PWSTR     ValueName,
    IN ULONG     ValueType,
    IN PVOID     ValueData,
    IN ULONG     ValueLength,
    IN PVOID     Context,
    IN PVOID     EntryContext
)
{

    PUCHAR       Buffer;

    if (ValueType != REG_MULTI_SZ) {

        return STATUS_OBJECT_NAME_NOT_FOUND;

    }

    Buffer=ExAllocatePool(NonPagedPool,ValueLength);

    if (Buffer==NULL) {

        return STATUS_INSUFFICIENT_RESOURCES;

    }

    RtlCopyMemory(
        Buffer,
        ValueData,
        ValueLength
        );

    *((PUCHAR *)EntryContext)=Buffer;

    return STATUS_SUCCESS;

}
