// Copyright(c) 1999 arton

#include "stdafx.h"
#include "MacCapv2.h"
#include "Vendors.h"

#include "NameValuePair.h"
#include "AssNum.h"

class CVendorEnum : public IEnumVARIANT
{
public:
	CVendorEnum(int iIndex = 0)
		: m_ulRef(1), m_iIndex(iIndex), m_iMax(GetVendorTableSize())
	{
	}
	~CVendorEnum()
	{
	}
	STDMETHOD_(ULONG, AddRef)()
	{
		return ++m_ulRef;
	}
	STDMETHOD_(ULONG, Release)()
	{
		m_ulRef--;
		if (m_ulRef <= 0)
		{
			delete this;
			return 0;
		}
		return m_ulRef;
	}
	STDMETHOD(QueryInterface)(REFIID iid, void ** ppvObject)
	{
		if (!ppvObject) return E_POINTER;
		if (IsEqualIID(IID_IUnknown, iid) 
			|| IsEqualIID(IID_IEnumVARIANT, iid)) {
			*ppvObject = this;
			m_ulRef++;
			return S_OK;
		}
		return E_NOINTERFACE;
	}

    HRESULT STDMETHODCALLTYPE Next(
		ULONG cElements, 
		VARIANT FAR* pvar, 
		ULONG FAR* pcElementFetched)
	{
		if (m_iIndex >= m_iMax) return S_FALSE;

		ULONG l2;
		if (pcElementFetched != NULL)        
			*pcElementFetched = 0;
		for (ULONG l=0; l<cElements; l++)        
			VariantInit(&pvar[l]);
		// Retrieve the next cElements elements.
		for (l2 = 0; m_iIndex < m_iMax && l2 < cElements; l2++, m_iIndex++)
		{
			const SVendor* set;
			for (; m_iIndex < m_iMax; m_iIndex++)
			{
				set = GetVendor(m_iIndex);
				if (set) break;
			}
			
			if (set == NULL)
				break;

			CComObject<CNameValuePair>* p;
			HRESULT hr = CComObject<CNameValuePair>::CreateInstance(&p);
			if (hr != S_OK)
				break;
			if (p->QueryInterface(IID_INameValuePair, (void**)&pvar[l2].pdispVal) != S_OK)
			{
				break;
			}
			p->m_strName = set->pVendor;
			p->m_value.vt = VT_I4;
			p->m_value.lVal = set->dwCode;
			pvar[l2].vt = VT_DISPATCH;
		}    
		// Set count of elements retrieved.
		if (pcElementFetched != NULL)        
			*pcElementFetched = l2;
		return  (l2 < cElements) ? ResultFromScode(S_FALSE) : ResultFromScode(S_OK);
	}
    HRESULT STDMETHODCALLTYPE Skip(unsigned long celt)
	{
		while (celt)
		{
			if (m_iIndex >= m_iMax) return S_FALSE;
			celt--;
			m_iIndex++;
		}
		return S_OK;
	}
    HRESULT STDMETHODCALLTYPE Reset()
	{
		m_iIndex = 0;
		return S_OK;
	}
    HRESULT STDMETHODCALLTYPE Clone(IEnumVARIANT FAR* FAR* ppenum)
	{
		if (!ppenum) return E_POINTER;
		*ppenum = new CVendorEnum(m_iIndex);
		return S_OK;
	}
private:
	ULONG m_ulRef;
	int m_iIndex;
	int m_iMax;
};

/////////////////////////////////////////////////////////////////////////////
// CVenders

HRESULT STDMETHODCALLTYPE CVendors::get__NewEnum( 
            /* [retval][out] */ IUnknown __RPC_FAR *__RPC_FAR *ppUnk)
{
	if (!ppUnk) return E_POINTER;
	*ppUnk = new CVendorEnum;
	return S_OK;
}
        
HRESULT STDMETHODCALLTYPE CVendors::get_Length( 
            /* [retval][out] */ long __RPC_FAR *pResult)
{
	if (!pResult) return E_POINTER;
	*pResult = GetVendorTableSize();
	return S_OK;
}
        
HRESULT STDMETHODCALLTYPE CVendors::Find( 
            /* [in] */ long vcode,
            /* [retval][out] */ BSTR __RPC_FAR *pResult)
{
	if (!pResult) return E_POINTER;
	BYTE b[6];
	ZeroMemory(b, sizeof(b));
	b[0] = (BYTE)(vcode >> 16);
	b[1] = (BYTE)(vcode >> 8);
	b[2] = (BYTE)vcode;
	const SVendor& v = SearchVendor(b);
	*pResult = SysAllocString(v.pVendor);
	return S_OK;
}

HRESULT STDMETHODCALLTYPE CVendors::get_Item( 
            /* [in] */ long index,
            /* [retval][out] */ INameValuePair __RPC_FAR *__RPC_FAR *ppRetVal)
{
	if (!ppRetVal) return E_POINTER;
	const SVendor* set = GetVendor(index);
	if (set == NULL) return DISP_E_BADINDEX;

	CComObject<CNameValuePair>* p;
	HRESULT hr = CComObject<CNameValuePair>::CreateInstance(&p);
	if (hr == S_OK)
	{
		hr = p->QueryInterface(IID_INameValuePair, (void**)ppRetVal);
		if (hr == S_OK)
		{
			p->m_strName = set->pVendor;
			p->m_value.vt = VT_I4;
			p->m_value.lVal = set->dwCode;
		}
	}
	return hr;
}

