// WrapperDispatch.h
// Copyright (c) 1999, Chris Sells
// All rights reserved. No warrenties extended.
/////////////////////////////////////////////////////////////////////////////
// The ATL Object Wizard generated an HTML Control that sets it's external
// dispatch with a simple static_cast:
//
//  hr = wnd.SetExternalDispatch(static_cast<IMyHtmlControlUI*>(this));
//
// This gives a non-default IDispatch to IE to use as the window.external
// objref. Under IE4, this works fine because IE4 never QIs this objref.
// However, IE5 *does* QI(IDispatch) on window.external, yielding the
// control's default IDispatch, e.g. IMyHtmlControl. The CWrapperDispatch
// class defined in this file provides a new COM identity to wrap around
// a non-default IDispatch so that QI will not yield another interface on the
// main object.
/////////////////////////////////////////////////////////////////////////////
// Known Issues:
// The WrapperDispatch object only implements IDispatch. It does so by
// forwarding calls to the non-default IDispatch that is provided at
// creation type. However, if you'd like to QI the window.external for any
// other interface that the control implements, you can't. The best way
// around this problem is to provide [propget] methods on the IMyHtmlControlUI
// interface that exposes other interfaces on the actual control. Or even
// provide a general purpose Object property for this purpose.
/////////////////////////////////////////////////////////////////////////////
// Usage:
/*
LRESULT CMyHtmlControl::OnCreate(UINT, WPARAM, LPARAM, BOOL&)
{
	CAxWindow wnd(m_hWnd);
	HRESULT hr = wnd.CreateControl(IDH_MYHTMLCONTROL);
    if( FAILED(hr) ) return hr;

    // Create a wrapper about the external dispatch interface
    CComObject<CWrapperDispatch>*   pdispWrapper;
    hr = pdispWrapper->CreateInstance(&pdispWrapper);
    if( FAILED(hr) ) return hr;

    // Set the external displatch interface
    CComPtr<IMyHtmlControlUI> pdispExternal = com_cast<IMyHtmlControlUI>(GetUnknown());
    if( !pdispExternal ) return E_UNEXPECTED;
    pdispWrapper->SetDispatch(pdispExternal);

	hr = wnd.SetExternalDispatch(pdispWrapper);
    if( FAILED(hr) ) return hr;

    hr = wnd.QueryControl(IID_IWebBrowser2, (void**)&m_spBrowser);
	return SUCCEEDED(hr) ? 0 : -1;
}
*/

#pragma once
#ifndef __WRAPPERDISPATCH_H__
#define __WRAPPERDISPATCH_H__

class CWrapperDispatch :
    public CComObjectRootEx<CComSingleThreadModel>,
    public CComCoClass<CWrapperDispatch>,
    public IDispatch
{
public:
    static
    HRESULT CreateDispatchWrapper(IDispatch* pdispToWrap, IDispatch** ppdispWrapper)
    {
        if( !ppdispWrapper ) return E_POINTER;
        *ppdispWrapper = 0;

        CComObject<CWrapperDispatch>*   pdispWrapper;
        HRESULT hr = pdispWrapper->CreateInstance(&pdispWrapper);
        if( FAILED(hr) ) return hr;
        
        pdispWrapper->SetDispatch(pdispToWrap);
        (*ppdispWrapper = pdispWrapper)->AddRef();
        return S_OK;
    }

public:
    void SetDispatch(IDispatch* pdisp)
    {
        m_spdisp = pdisp;   // NOTE: No QI
    }

DECLARE_NO_REGISTRY()
BEGIN_COM_MAP(CWrapperDispatch)
    COM_INTERFACE_ENTRY(IDispatch)
END_COM_MAP()

// IDispatch
    STDMETHODIMP GetTypeInfoCount(UINT* pctinfo)
    {
        if( !m_spdisp ) return E_UNEXPECTED;
        return m_spdisp->GetTypeInfoCount(pctinfo);
    }

	STDMETHODIMP GetTypeInfo(UINT itinfo, LCID lcid, ITypeInfo** pptinfo)
    {
        if( !pptinfo ) return E_POINTER;
        *pptinfo = 0;
        if( !m_spdisp ) return E_UNEXPECTED;
        return m_spdisp->GetTypeInfo(itinfo, lcid, pptinfo);
    }

	STDMETHODIMP GetIDsOfNames(REFIID riid, LPOLESTR* rgszNames, UINT cNames, LCID lcid, DISPID* rgdispid)
    {
        if( !m_spdisp ) return E_UNEXPECTED;
        return m_spdisp->GetIDsOfNames(riid, rgszNames, cNames, lcid, rgdispid);
    }

	STDMETHODIMP Invoke(DISPID dispidMember, REFIID riid, LCID lcid, WORD wFlags, DISPPARAMS* pdispparams, VARIANT* pvarResult, EXCEPINFO* pexcepinfo, UINT* puArgErr)
    {
        if( !m_spdisp ) return E_UNEXPECTED;
        return m_spdisp->Invoke(dispidMember, riid, lcid, wFlags, pdispparams, pvarResult, pexcepinfo, puArgErr);
    }

private:
    CComPtr<IDispatch>  m_spdisp;
};

#endif  // __WRAPPERDISPATCH_H__
